# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/BioFusion',
                'doc_host': 'https://CalmScout.github.io',
                'git_url': 'https://github.com/CalmScout/BioFusion',
                'lib_path': 'BioFusion'},
  'syms': { 'BioFusion.cmmd': { 'BioFusion.cmmd.cmmd': ('cmmd.html#cmmd', 'BioFusion/cmmd.py'),
                                'BioFusion.cmmd.communities_detection': ('cmmd.html#communities_detection', 'BioFusion/cmmd.py'),
                                'BioFusion.cmmd.community_detection_ray': ('cmmd.html#community_detection_ray', 'BioFusion/cmmd.py'),
                                'BioFusion.cmmd.continual_multiplex_analysis': ( 'cmmd.html#continual_multiplex_analysis',
                                                                                 'BioFusion/cmmd.py')},
            'BioFusion.download_data': { 'BioFusion.download_data.create_dataframe_from_edges': ( 'download_data.html#create_dataframe_from_edges',
                                                                                                  'BioFusion/download_data.py')},
            'BioFusion.utils': { 'BioFusion.utils.create_and_save_random_graph': ( 'synthetic_data.html#create_and_save_random_graph',
                                                                                   'BioFusion/utils.py'),
                                 'BioFusion.utils.create_random_graph': ('synthetic_data.html#create_random_graph', 'BioFusion/utils.py'),
                                 'BioFusion.utils.generate_and_save_graphs': ( 'synthetic_data.html#generate_and_save_graphs',
                                                                               'BioFusion/utils.py'),
                                 'BioFusion.utils.obfuscate_nodes': ('synthetic_data.html#obfuscate_nodes', 'BioFusion/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_synthetic_data.ipynb.

# %% auto 0
__all__ = ['obfuscate_nodes', 'create_random_graph', 'create_and_save_random_graph', 'generate_and_save_graphs']

# %% ../nbs/01_synthetic_data.ipynb 5
import networkx as nx
import csv
from pathlib import Path
import random
import string
from tqdm import tqdm
from pympler.asizeof import asizeof

# %% ../nbs/01_synthetic_data.ipynb 9
def obfuscate_nodes(n_nodes:int,
                    node_labels_pool:list[str]=None,
                    str_len:int=10) -> dict:
    """
    Maps nodes from range [0, n_nodes) to unique random strings of length str_len.
    All upper case letters and digits are used, starts with a letter.
    If node_labels_pool is provided, nodes selected from it if len(node_labels_pool) >= n_nodes,
    otherwise, additional unique nodes are generated.
    """
    if node_labels_pool is not None and len(node_labels_pool) > 0:
        assert len(node_labels_pool[0]) == str_len
    
    def encode(str_len):
        return ''.join(random.choices(string.ascii_uppercase, k=1)) +\
               ''.join(random.choices(string.ascii_uppercase + string.digits, k=(str_len-1)))
    
    if node_labels_pool is not None:
        if len(node_labels_pool) >= n_nodes:
            # select random nodes from node_labels_pool
            random.shuffle(node_labels_pool)
            vals = node_labels_pool[:n_nodes]
            d = dict(zip(range(n_nodes), vals))
        elif len(node_labels_pool) < n_nodes:
            # use all nodes from node_labels_pool
            random.shuffle(node_labels_pool)
            vals = node_labels_pool[:]
            # generate additional unique nodes
            for i in range(n_nodes - len(node_labels_pool)):
                x = encode(str_len)
                while x in vals: # guarantee uniqueness
                    x = encode(str_len)
                vals.append(x)
            # shuffle pre-defined and newly generated nodes
            random.shuffle(vals)
            d = dict(zip(range(n_nodes), vals))
    else:   # no node_labels_pool provided
        d = {}
        for i in range(n_nodes):
            x = encode(str_len)
            # gaurantee uniqueness
            while x in d.values():
                x = encode(str_len)
            
            d[i] = x
    
    return d

# %% ../nbs/01_synthetic_data.ipynb 14
def create_random_graph(generator:callable, obfuscate:bool=True, node_labels_pool:list[str]=None, **kwargs):
    """Thin wrapper around networkx's random graph generator

    Args:
        generator (callable): networkx's random graph generator
        **kwargs: passed to generator
    Example:
        >>> G = create_random_graph(nx.erdos_renyi_graph, n=10, p=0.6, directed=False)
    """
    G = generator(**kwargs)
    if obfuscate:
        G = nx.relabel_nodes(G, obfuscate_nodes(G.number_of_nodes(), node_labels_pool=node_labels_pool))
    return G

# %% ../nbs/01_synthetic_data.ipynb 19
def create_and_save_random_graph(generator:callable,
                                 label_edges:str,
                                 path_to:str|Path,
                                 obfuscate:bool=True,
                                 node_labels_pool:list[str]=None,
                                 **kwargs) -> nx.classes.graph.Graph:
    """
    Creates a random graph and saves it to path_to.

    Args:
        generator (callable): networkx's random graph generator
        obfuscate (bool, optional): if true, obfuscates node names. Defaults to True.
        path_to (str|Path, optional): path to save the graph. Defaults to None.
    
    Example:
        >>> create_random_graph_and_save(nx.erdos_renyi_graph, n=10, p=0.6, directed=False, path_to=path_data / 'synthetic' / '1.csv')
    """
    path_to = Path(path_to)

    G = create_random_graph(generator=generator, obfuscate=obfuscate, node_labels_pool=node_labels_pool, **kwargs)
    edges = G.edges()
    with open(path_to, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=' ')
        for edge in edges:
            writer.writerow([edge[0], edge[1], label_edges])
    
    return G

# %% ../nbs/01_synthetic_data.ipynb 26
def generate_and_save_graphs(graph_params, path_dir_to):
    """
    Generates and saves multiple random graphs based on provided parameters.

    Parameters:
    graph_params (list of tuples): A list of tuples where each tuple contains:
        - n (int): Number of nodes in the graph.
        - p (float): Probability of edge creation.
        - label_edges (str): Label for the edges.
    path_dir_to (Path): Directory path where the graph files will be saved.
    
    The method iterates over the provided graph parameters, generates an Erdős–Rényi graph,
    saves it to a CSV file, and maintains a pool of node labels to ensure uniqueness.
    A progress bar is displayed to indicate the number of layers being processed.
    """
    path_dir_to = Path(path_dir_to)
    node_labels_pool = []

    graphs = []
    
    for i, (n, p, label_edges) in enumerate(tqdm(graph_params, desc='Generating layers'), start=1):
        G = create_and_save_random_graph(nx.erdos_renyi_graph, label_edges, path_dir_to / f'{i}.csv',
                                         obfuscate=True, node_labels_pool=node_labels_pool, n=n, p=p)
        node_labels_pool.extend(list(G.nodes()))

        graphs.append(G)
    
    # collect metrics of graphs
    total_nodes = sum(G.number_of_nodes() for G in graphs)
    total_edges = sum(G.number_of_edges() for G in graphs)
    total_memory = sum(asizeof(G) for G in graphs) / (1024 ** 2)  # Approximate memory usage, MBs

    return (total_nodes, total_edges, total_memory)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_cmmd.ipynb.

# %% auto 0
__all__ = ['community_detection_ray', 'communities_detection', 'continual_multiplex_analysis', 'cmmd']

# %% ../nbs/03_cmmd.ipynb 4
import os
import sys
import subprocess
import numpy as np
from concurrent.futures import ThreadPoolExecutor
from scipy.spatial.distance import pdist, squareform
import seaborn as sns
import pickle
from pathlib import Path
import pathlib
import time
import shutil
import ray

# %% ../nbs/03_cmmd.ipynb 10
@ray.remote
def community_detection_ray(method, current_destfile, current_resolution, layers):
    system_order = f"../bin/{method} -o {current_destfile} -p {current_resolution} {layers} > /dev/null"
    subprocess.run(system_order, shell=True)

def communities_detection(input_layers: list[str] = None,
                          gamma_min: float = None,
                          gamma_max: float = None,
                          gamma_step: float = None,
                          path_to_communities: str = None,
                          method: str = "molti",
                          batch_size: int = 10,
                          num_cpus: int = 1):
    assert method in ["molti", "molti-dream"]
    layers = " ".join(input_layers)
    resolution_gamma_step = np.arange(gamma_min, gamma_max + gamma_step, gamma_step)
    desfile_vector = [f"{path_to_communities}{res}.csv" for res in resolution_gamma_step]

    tasks = []
    for i, current_resolution in enumerate(resolution_gamma_step):
        current_destfile = desfile_vector[i]
        task = community_detection_ray.options(num_cpus=num_cpus).remote(method, current_destfile, current_resolution, layers)
        tasks.append(task)

        # Execute tasks in batches
        if (i + 1) % batch_size == 0 or i == len(resolution_gamma_step) - 1:
            ray.get(tasks)  # Wait for the current batch to complete
            tasks = []  # Reset the task list for the next batch

    # Ensure any remaining tasks are completed
    if tasks:
        ray.get(tasks)

# %% ../nbs/03_cmmd.ipynb 11
def continual_multiplex_analysis(nodelist:list[str]=None,
                                 path_to_communities:str=None,
                                 distmethod:str="hamming"):
        # reading MolTi output files
        output_files = [f for f in os.listdir(path_to_communities) if "_" not in f]
        
        alllists = []
        
        for output_file in output_files:
            with open(os.path.join(path_to_communities, output_file), 'r') as file:
                red = file.readlines()
            
            cluster_ids = [i for i, line in enumerate(red) if "Cluster" in line]
            lista = []
            
            for j, st in enumerate(cluster_ids):
                if j == len(cluster_ids) - 1:
                    en = len(red)
                else:
                    en = cluster_ids[j + 1]
                current_cluster = red[st:en]
                current_cluster2 = current_cluster[:-2] if j != len(cluster_ids) - 1 else current_cluster[:-1]
                lista.append(current_cluster2[1:])
            
            alllists.append(lista)
        
        allgenes = list(set([gene for sublist in alllists for cluster in sublist for gene in cluster]))

        if nodelist:
            allgenes = list(set(allgenes).intersection(nodelist))
        
        # Calculating Gene/Community matrix
        res_matrix = np.empty((len(allgenes), len(alllists) + 1), dtype=object)
        res_matrix[:, :-1] = 0  # Initialize the integer part of the matrix with zeros
        gene_indices = {gene: idx for idx, gene in enumerate(allgenes)}
        
        for j, output_file_list in enumerate(alllists):
            for k, cluster in enumerate(output_file_list):
                for gene in cluster:
                    res_matrix[gene_indices[gene], j] = k + 1
        
        patterns = ["_".join(map(str, res_matrix[i, :-1])) for i in range(len(allgenes))]
        res_matrix[:, -1] = np.array(patterns, dtype=str)
        
        # Calculating Hamming distances for all gene pairs
        gene_community_matrix = res_matrix[:, :-1].astype(np.int64)
        genes_same_communities = {pattern: [] for pattern in np.unique(gene_community_matrix[:,-1])}
        
        for i, pattern in enumerate(gene_community_matrix[:,-1]):
            genes_same_communities[pattern].append(allgenes[i])
        
        with ThreadPoolExecutor(max_workers=os.cpu_count()) as executor:
            distance_matrix = squareform(pdist(gene_community_matrix, metric=distmethod))
        
        final_output = {
            "gene_community_matrix": gene_community_matrix,
            "l_constant": genes_same_communities,
            "distance_matrix": distance_matrix
        }
        
        return final_output

# %% ../nbs/03_cmmd.ipynb 12
def cmmd(nodelist:list[str]|None=None,
         input_layers:list[str]=None,
         gamma_min:float=None,
         gamma_max:float=None,
         gamma_step:float=None, 
         distmethod:str="hamming",
         method:str="molti",
         path_to_communities:str=None,
         batch_size: int = 10,
         num_cpus: int = 1):
    
    """
    Compute CmmD multilayer community trajectory analysis for a set of given networks.

    Parameters
    ----------
    nodelist : list, optional
        A list with the unique nodes that we want to appear in the final output. If not given,
        all nodes of the multiplex will be in the final output (nodelist= NULL)
    input_layers : list
        A vector of strings containing the paths where the different network layers are located
        in the system. Networks should be a two column file representing the edges of the graph.
    gamma_min : float
        The first gamma resolution parameter to use in the different MolTi's analysis
    gamma_max : float
        The last gamma resolution parameter to use in the different MolTi's analysis.
    gamma_step : float
        The gamma_step of the resolution parameter to use. 
    distmethod : str, optional
        A distance method metric to use to compute the trajectories. Defaults to "hamming" for hamming
        distance, but accepts any other metric supplied by scipy.spatial.distance.pdist.
    path_to_communities : str, optional
        The path to save Molti's output files. Defaults to "Output/".

    Returns
    -------
    A dictionary containing the following keys:
        gene_community_matrix: A matrix where the rows correspond to the different genes, and the columns to the different community structures. The values of the matrix are the cluster to which the gene belongs in the corresponding community structure.
        l_constant: A dictionary where the keys are the different community structures, and the values are the list of genes that belong to that community structure.
        distance_matrix: A matrix with the hamming distances between all pairs of genes.
    """
    # 0. check input correctness
    if input_layers is None or len(input_layers) < 1:
        raise ValueError("ERROR: Input_layers argument must be a list of at least 1 network file")
    
    if not isinstance(gamma_max, (int, float)):
        raise ValueError("ERROR: Resolution parameter must be a number")
    
    if not isinstance(gamma_min, (int, float)):
        raise ValueError("ERROR: Resolution parameter must be a number")
    
    if not isinstance(gamma_step, (int, float)):
        raise ValueError("ERROR: gamma_step value must be a number")
    
    if not isinstance(path_to_communities, str):
        raise ValueError("ERROR: path_to_communities expects a character string")
    
    assert distmethod in ['braycurtis', 'canberra', 'chebyshev', 'cityblock',
                          'correlation', 'cosine', 'dice', 'euclidean', 'hamming',
                          'jaccard', 'jensenshannon', 'kulczynski1', 'mahalanobis',
                          'matching', 'minkowski', 'rogerstanimoto', 'russellrao',
                          'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean',
                          'yule']

    # 1st part: community detection
    Path(path_to_communities).mkdir(parents=True, exist_ok=True)
    # if folder is empty, generate communities to populate it
    if len(os.listdir(path_to_communities)) == 0:
        communities_detection(input_layers=input_layers, gamma_min=gamma_min, gamma_max=gamma_max,
                            gamma_step=gamma_step, method=method, path_to_communities=path_to_communities,
                            batch_size=batch_size, num_cpus=num_cpus)
    
    # 2nd part: cmmd
    final_output = continual_multiplex_analysis(nodelist=nodelist,
                                                path_to_communities=path_to_communities,
                                                distmethod=distmethod)
    
    return final_output



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_download_data.ipynb.

# %% auto 0
__all__ = ['create_dataframe_from_edges']

# %% ../nbs/02_download_data.ipynb 4
import requests
import zipfile
from datetime import datetime
from tqdm import tqdm
import json
from pathlib import Path
import os
from itertools import combinations
import re
import pandas as pd
import ast
import rdflib
import multiprocessing
from joblib import Parallel, delayed
import gzip
import shutil
import cobra
import urllib.request
from collections import defaultdict

from bs4 import BeautifulSoup

# %% ../nbs/02_download_data.ipynb 17
def create_dataframe_from_edges(data):
    """
    Creates a DataFrame from a dictionary of edges, dynamically extracting attributes as columns,
    formatting column names to lowercase with underscores replacing spaces,
    and replacing spaces in all column values with underscores.

    Parameters:
        data (dict): A dictionary containing an 'edges' key with a list of edge dictionaries.

    Returns:
        pd.DataFrame: A DataFrame representing the edges with dynamically extracted attributes.
    """
    # Extract all unique attribute keys from 'attributes'
    attribute_keys = set()
    for item in data.get('edges', []):
        if 'attributes' in item and isinstance(item['attributes'], dict):
            attribute_keys.update(item['attributes'].keys())

    # Normalize attribute keys
    normalized_keys = {key: key.lower().replace(" ", "_") for key in attribute_keys}

    # Create the DataFrame
    df_edges = pd.DataFrame([
        {
            'source': item.get('source'),
            'target': item.get('target'),
            **{normalized_keys[key]: item.get('attributes', {}).get(key) for key in attribute_keys}
        }
        for item in data.get('edges', [])
    ])

    # Normalize column names for source and target
    df_edges.rename(columns=lambda col: col.lower().replace(" ", "_"), inplace=True)

    # Replace spaces with underscores in all column values
    df_edges = df_edges.map(lambda x: x.replace(" ", "_") if isinstance(x, str) else x)

    return df_edges

